# Exercices on Chapter 10 (neural networks)
library(nnet)

# Exercise 1: regression

library(MASS)  
n<-nrow(Boston)

#Boston<-scale(Boston)

ntrain<-round(2*n/3)
ntest<-n-ntrain
train<-sample(n,ntrain)
Boston.train<-Boston[train,]
Boston.test<-Boston[-train,]
y.test<-Boston.test$medv

# Experiment with NN of different sizes
ntrials<-10
best.error<-Inf
for(i in 1:ntrials){
  fit<-nnet(medv ~ lstat,data=Boston.train,linout=TRUE,size=3)
  if(fit$value<best.error){
    best.error<-fit$value
    best.fit<-fit
  }
}

ypred<-predict(best.fit,newdata=Boston.test)
plot(y.test,ypred)
abline(0,1)
print(mean((y.test-ypred)^2))


plot(Boston$lstat,Boston$medv)
xx<-seq(min(Boston$lstat),max(Boston$lstat),0.1)
yy<-predict(best.fit,newdata=data.frame(lstat=xx))
lines(xx,yy,col='red',lwd=2)

# Comparison with smoothing spline
library(splines)

fit.ss=smooth.spline(Boston.train$lstat,Boston.train$medv,cv=TRUE)

pred<-predict(fit.ss,newdata=data.frame(lstat=xx))
lines(pred$x,pred$y,col='blue',lwd=2)

pred.ss<-predict(fit.ss,newdata=Boston.test)
print(mean((y.test-pred.ss$y)^2))


# Cross-validation

K<-10
folds=sample(1:K,ntrain,replace=TRUE)
Lambda<-c(1e-6,1e-5,1e-4,0.001,0.01,0.1,1,10,100,1000)
N<-length(Lambda)
CV1<-rep(0,N)
for(i in (1:N)){
  print(i)
  for(k in (1:K)){
    fit=nnet(medv~lstat,data=Boston.train[folds!=k,],
             size=10,decay=Lambda[i],lin=TRUE,trace=FALSE)
    pred<-predict(fit,newdata=Boston.train[folds==k,])
    CV1[i]<-CV1[i]+ sum((Boston.train$medv[folds==k]-pred)^2)
  }
  CV1[i]<-CV1[i]/ntrain
}
plot(Lambda,CV1,type='b',xlab=expression(lambda),
     ylab='CV error',log="x")

best.fit=nnet(medv~lstat,data=Boston.train,
         size=10,decay=1,lin=TRUE,trace=FALSE)

ypred<-predict(best.fit,newdata=Boston.test)
plot(y.test,ypred)
abline(0,1)
print(mean((y.test-ypred)^2))


plot(Boston$lstat,Boston$medv)
xx<-seq(min(Boston$lstat),max(Boston$lstat),0.1)
yy<-predict(best.fit,newdata=data.frame(lstat=xx))
lines(xx,yy,col='red',lwd=2)


# Train with all variables
# We scale the predictors first

x<-Boston[,-14]
y<-Boston[,14]
x<-scale(x)
x.train<-x[train,]
y.train<-y[train]
x.test<-x[-train,]
y.test<-y[-train]

# Cross-validation

K<-10
folds=sample(1:K,ntrain,replace=TRUE)
Lambda<-c(1e-6,1e-5,1e-4,0.001,0.01,0.1,1,10,100,1000)
N<-length(Lambda)
CV1<-rep(0,N)
for(i in (1:N)){
  print(i)
  for(k in (1:K)){
    ntrials<-10
    best.error<-Inf
    for(j in 1:ntrials){
#      fit=nnet(medv~.,data=Boston.train[folds!=k,],
#               size=10,decay=Lambda[i],lin=TRUE,trace=FALSE)
      fit=nnet(x.train[folds!=k,],y.train[folds!=k],
               size=10,decay=Lambda[i],lin=TRUE,trace=FALSE)
      if(fit$value<best.error){
        best.error<-fit$value
        best.fit<-fit
      }
    }
#    pred<-predict(best.fit,newdata=Boston.train[folds==k,])
    pred<-predict(best.fit,newdata=x.train[folds==k,])
#    CV1[i]<-CV1[i]+ sum((Boston.train$medv[folds==k]-pred)^2)
    CV1[i]<-CV1[i]+ sum((y.train[folds==k]-pred)^2)
  }
  CV1[i]<-CV1[i]/ntrain
}
plot(Lambda,CV1,type='b',xlab=expression(lambda),
     ylab='CV error',log="x")


ntrials<-20
best.error<-Inf
for(i in 1:ntrials){
  fit<-nnet(x.train,y.train,linout=TRUE,size=10,
            decay=10,trace=FALSE)
  if(fit$value<best.error){
    best.error<-fit$value
    best.fit<-fit
  }
}

ypred<-predict(best.fit,newdata=x.test)
plot(y.test,ypred)
abline(0,1)
print(mean((y.test-ypred)^2))

# Comparison with random forest

library(randomForest)
fit.rf<-randomForest(medv~.,data=Boston.train)
pred.rf<-predict(fit.rf,newdata=Boston.test)
mse.rf<-mean((y.test-pred.rf)^2)
print(mse.rf)
plot(y.test,pred.rf)


# ----------------------------------------------
# Exercise 2
# Time series data

data<-read.csv("/Users/Thierry/Documents/R/Data/Economics/CP.csv",header=TRUE)
x<-ts(data$Cp,start = c(1959,1),frequency=12)
plot(x)

y<-diff(log(x))
plot(y)

preproc<-function(y,W,H){
  n<-length(y)
  Lag<-W+H
  Z<-y[1:(n-Lag+1)]
  for(i in 2:Lag) Z<-cbind(Z,y[i:(n-Lag+i)])
  X<-Z[,1:W]
  Y<-Z[,(W+1):(W+H)]
  return(list(X=X,Y=Y))
}

data1<-preproc(y,W=2,H=1)
X<-data1$X
Xs<-scale(X)
Y<-data1$Y

n<-nrow(X)
ntrain<-n-174
X.train<-X[1:ntrain,]
Xs.train<-Xs[1:ntrain,]
Y.train<-Y[1:ntrain]
X.test<-X[(ntrain+1):n,]
Xs.test<-Xs[(ntrain+1):n,]
Y.test<-Y[(ntrain+1):n]


# Cross-validation


K<-10
folds=sample(1:K,ntrain,replace=TRUE)
Lambda<-c(1e-6,1e-5,1e-4,0.001,0.01,0.1,1,10,100,1000)
N<-length(Lambda)
CV1<-rep(0,N)
for(i in (1:N)){
  print(i)
  for(k in (1:K)){
    fit<-nnet(Xs.train[folds!=k,],Y.train[folds!=k], linout = TRUE,size=5,
              decay=Lambda[i],maxit = 500,trace=FALSE)
    pred<-predict(fit,newdata=Xs.train[folds==k,])
    CV1[i]<-CV1[i]+ sum((Y.train[folds==k]-pred)^2)
  }
  CV1[i]<-CV1[i]/ntrain
}
plot(Lambda,CV1,type='b',xlab=expression(lambda),
     ylab='CV error',log="x")



fit<-nnet(Xs.train,Y.train, linout = TRUE,size=5,decay=0.1,
          maxit = 500)
y.pred<-predict(fit,newdata=Xs.test)
plot(Y.test,y.pred)
abline(0,1)

plot(Y.test,type="l")
lines(y.pred,col="red")
     
err.NN<-mean((Y.test-y.pred)^2)
print(err.NN)

pred.naive<-X.test[,2]
err.naive<-mean((Y.test-pred.naive)^2)
print(err.naive)

# AR-GARCH model
library(rugarch)
y.test<-window(y,start=c(1984,2))


model<-ugarchspec(variance.model = list(model = "sGARCH", garchOrder = c(1, 1)), 
                  mean.model = list(armaOrder = c(2, 1), include.mean = TRUE), 
                  distribution.model = "norm")

fit.garch<-ugarchfit(spec=model, data=y, out.sample = 174, solver = "solnp", 
                     #solver.control = list(),
                     fit.control = list(stationarity = 1, fixed.se = 0, 
                                        scale = 0, rec.init = 'all',
                                        trunclag = 1000),
                     numderiv.control = list(grad.eps=1e-4, grad.d=0.0001,
                                             grad.zero.tol=sqrt(.Machine$double.eps/7e-7), hess.eps=1e-4, hess.d=0.1,
                                             hess.zero.tol=sqrt(.Machine$double.eps/7e-7), r=4, v=2))


pred.garch<-ugarchforecast(fitORspec=fit.garch,n.ahead = 3, n.roll = 173)
y.pred.garch<-t(fitted(pred.garch))

mean((y.test-y.pred.garch[,1])^2)


plot(as.vector(y.test),type="l")
lines(y.pred.garch[,1],col="blue")
lines(y.pred,col="red")


